"""
HTMD provides a variety of protocols for setting up different types of simulations.
The most basic protocols include equilibration and production runs, both of which support flat-bottom potentials and
constraints for atoms.
"""
# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
__author__ = 'gianni'
