from moleculekit.projections.metricsecondarystructure import MetricSecondaryStructure
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricsecondarystructure. This will be deprecated. '
                'Use moleculekit.projections.metricsecondarystructure instead.')