from moleculekit.projections.metricshell import MetricShell
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricshell. This will be deprecated. '
                'Use moleculekit.projections.metricshell instead.')