from moleculekit.projections.metricdihedral import *
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricdihedral. This will be deprecated. '
                'Use moleculekit.projections.metricdihedral instead.')