from moleculekit.projections.metricsphericalcoordinate import MetricSphericalCoordinate
import logging


logger = logging.getLogger(__name__)

logger.warning('Please do not import from htmd.projections.metricsphericalcoordinate. This will be deprecated. '
                'Use moleculekit.projections.metricsphericalcoordinate instead.')